/*
 * loadgen.c	: generate schema for benchw suite
 * 
 * by Mark Kirkwood (markir@paradise.net.nz)
 *
 * Bugs 		: should really check return code from all malloc's
 */


/*
 * required includes
 */
#include "inc.h"


/*
 * relation structures
 */
#include "rel.h"


/*
 * lod specific defines
 */
#include "lod.h"


/*
 * show nice description of a database engine type
 */
char	*db_name(int db_type) {

	switch (db_type) {
		case PG :
			return "Postgresql";
			break;
		case MY :
			return "Mysql";
			break;
		case DB2 :
			return "DB2";
			break;
		case ORA :
			return "Oracle";
			break;
		case SAP :
			return "Sapdb";
			break;
		case FB :
			return "Firebird";
			break;
		default :
			return "Unknown";
			break;
	}
}


/*
 *  write load file for Postgresql
 */
int		write_lod_pg(char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i;
	int		fd;


	/*
	 * work out the file name
	 */
	lodfile = malloc(strlen(dpath) + 10);
	sprintf(lodfile, "%s%s%s", dpath, "/", LODFILE);

	/*
	 * open the file
	 */
	if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("        :    %s\n", lodfile);



	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);


	/*
	 * start writing the buffer out
	 *
	 */
	sprintf(buffer, "--\n--  benchw load : generated by loadgen\n--              : Postgresql variant\n--\n");


	/*
	 * get the relation structure
	 */
	rels = getrels();



	/*
	 * just use the relation name
	 * then write out the buffer
	 */
	for ( i = 0; i < NORELS ; i++ ) {
		sprintf(tempbuf, "COPY %s FROM '%s/%s.dat' \nUSING DELIMITERS ',';\n\n",
						rels[i].name, lpath, rels[i].name );
		strcat(buffer, tempbuf);
	}


	if( write(fd, buffer, strlen(buffer)) == -1 ) {
		perror("failed to write");
		return 2;
	}


	freerels(rels);

	free(tempbuf);
	free(buffer);

	close(fd);
	free(lodfile);

	return 0;
}


/*
 *  write load file for DB2
 */
int		write_lod_db2(char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i,j;
	int		hasdate;
	int		fd;


	/*
	 * work out the file name
	 */
	lodfile = malloc(strlen(dpath) + 10);
	sprintf(lodfile, "%s%s%s", dpath, "/", LODFILE);

	/*
	 * open the file
	 */
	if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("        :    %s\n", lodfile);



	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);


	/*
	 * start writing the buffer out
	 *
	 */
	sprintf(buffer, "--\n--  benchw load : generated by loadgen\n--              : DB2 variant\n--\n");


	/*
	 * get the relation structure
	 */
	rels = getrels();



	/*
	 * just use the relation name
	 * then write out the buffer
	 */

	for ( i = 0; i < NORELS ; i++ ) {
		sprintf(tempbuf, "LOAD FROM '%s%s%s.dat' OF DEL \n",
						lpath, "/", rels[i].name);
		/*
		 * check attributes for date type
		 */
		hasdate = 0;
		for (j = 0; j < rels[i].noattrs; j++) {
			if ( strcmp(rels[i].attrs[j].type, "DATE") == 0) {
				hasdate = 1;
			}
		}

		if (hasdate) {
			strcat(tempbuf, "MODIFIED BY TIMESTAMPFORMAT=\"YYYY-MM-DD\"\n" );
		}
		strcat(buffer, tempbuf);
		sprintf(tempbuf, "INSERT INTO %s;\n\n", rels[i].name );
		strcat(buffer, tempbuf);
	}


	if( write(fd, buffer, strlen(buffer)) == -1 ) {
		perror("failed to write");
		return 2;
	}


	freerels(rels);

	free(tempbuf);
	free(buffer);

	close(fd);
	free(lodfile);

	return 0;
}


/*
 *  write load file for Mysql
 */
int		write_lod_my( char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i;
	int		fd;


	/*
	 * work out the file name
	 */
	lodfile = malloc(strlen(dpath) + 10);
	sprintf(lodfile, "%s%s%s", dpath, "/", LODFILE);

	/*
	 * open the file
	 */
	if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("        :    %s\n", lodfile);


	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);


	/*
	 * start writing the buffer out
	 *
	 */
	sprintf(buffer, "--\n--  benchw load : generated by loadgen\n--              : Mysql variant\n--\n");


	/*
	 * get the relation structure
	 */
	rels = getrels();



	/*
	 * just use the relation name
	 * then write out the buffer
	 */
	for ( i = 0; i < NORELS ; i++ ) {
		sprintf(tempbuf, "LOAD DATA LOCAL INFILE '%s/%s.dat' \nINTO TABLE %s FIELDS TERMINATED BY ',';\n\n",
						lpath, rels[i].name, rels[i].name );
		strcat(buffer, tempbuf);
	}


	if( write(fd, buffer, strlen(buffer)) == -1 ) {
		perror("failed to write");
		return 2;
	}


	freerels(rels);
	
	free(tempbuf);
	free(buffer);

	close(fd);
	free(lodfile);

	return 0;
}


/*
 *  write load control files for Oracle
 */
int		write_lod_ora(char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i,j;
	int		fd;


	printf("        : control files for sqlldr\n");

	/*
	 *  allocate enough space for any of the filenames
	 */
	lodfile = malloc(strlen(dpath) + IDSIZ + 10);


	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);



	/*
	 * get the relation structure
	 */
	rels = getrels();


	/*
	 * open a new file for each relation
	 *
	 */
	for ( i = 0; i < NORELS ; i++ ) {

		sprintf(lodfile, "%s%s%s%s", dpath, "/", rels[i].name, ".ctl");

		if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
			perror("could not create");
			return 1;
		}
		printf("        :    %s\n", lodfile);


		sprintf(buffer, 
				"LOAD DATA \nINFILE \"%s%s%s.dat\" \nINTO TABLE %s \nFIELDS TERMINATED BY ',' TRAILING NULLCOLS\n(",
 				lpath, "/", rels[i].name, rels[i].name );
		/*
		 * now include the attributes
		 * catch date datatypes and write a format mask
		 */
		for (j = 0; j < rels[i].noattrs; j++) {
			sprintf(tempbuf, "\t%s", rels[i].attrs[j].name);
			if ( strcmp(rels[i].attrs[j].type, "DATE") == 0) {
				strcat(tempbuf, " DATE \"YYYY-MM-DD\"");
			}

			if ( j != rels[i].noattrs -1 ) {		// comma if not at end of attrs !
				strcat(tempbuf, ",");
			}
			strcat(tempbuf, "\n");
			strcat(buffer, tempbuf);
		}
		strcat(buffer,")\n");

		/*
		 * write it out
		 */
		if( write(fd, buffer, strlen(buffer)) == -1 ) {
			perror("failed to write");
			return 2;
		}

		close(fd);
	}


	freerels(rels);
	
	free(tempbuf);
	free(buffer);
	free(lodfile);

	return 0;
}


/*
 *  write load control files for Sapdb (follow Oracle style)
 */
int		write_lod_sap(char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i;
	int		fd;


	printf("        : script files for repmcli\n");

	/*
	 *  allocate enough space for any of the filenames
	 */
	lodfile = malloc(strlen(dpath) + IDSIZ + 10);


	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);



	/*
	 * get the relation structure
	 */
	rels = getrels();


	/*
	 * open a new file for each relation
	 *
	 */
	for ( i = 0; i < NORELS ; i++ ) {

		sprintf(lodfile, "%s%s%s%s", dpath, "/", rels[i].name, ".rep");

		if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
			perror("could not create");
			return 1;
		}
		printf("        :    %s\n", lodfile);



		sprintf(buffer, "FASTLOAD \nTABLE %s\nINFILE '%s%s%s.dat'\n",
				rels[i].name, lpath, "/", rels[i].name );
		strcat(buffer, "SEPARATOR ',' DELIMITER ''\n");
		if ( strcmp(rels[i].type, "timelike dimension") == 0) {
			strcat(buffer, "DATE 'YYYY-MM-DD'\n");
		}

		/*
		 * write it out
		 */
		if( write(fd, buffer, strlen(buffer)) == -1 ) {
			perror("failed to write");
			return 2;
		}

		close(fd);
	}


	freerels(rels);

	free(tempbuf);
	free(buffer);
	free(lodfile);

	return 0;
}


/*
 *  write load control file for Firebird 
 */
int		write_lod_fb(char *lpath, char *dpath) {

	Rel		*rels;
	char	*lodfile;
	char	*buffer;
	char	*tempbuf;
	int		i, j;
	int		fd;


	printf("        : script files for external tables\n");


	/*
	 * get the relation structure
	 */
	rels = getrels();


	/*
	 * work out the file name
	 */
	lodfile = malloc(strlen(dpath) + 10);
	sprintf(lodfile, "%s%s%s", dpath, "/", LODFILE);

	/*
	 * open the file
	 */
	if ((fd = open(lodfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("        :    %s\n", lodfile);

	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(8192);
	tempbuf = malloc(512);

	sprintf(buffer, "--\n--  benchw load : generated by loadgen\n--              : Firebird variant\n--\n");


	/*
	 * loop through the relations
	 * and their attributes -
	 * firstly creating an external table for each (remember the newline)
	 * and then inserting from this into the real one.
	 */
	for ( i = 0; i < NORELS ; i++ ) {

		/* create the external table */
		sprintf(buffer, "CREATE TABLE %s_ext EXTERNAL FILE '%s%s%s.dat' (\n",
				rels[i].name, lpath, "/", rels[i].name );

		for (j = 0; j < rels[i].noattrs; j++) {
			sprintf(tempbuf, "\t%-30s CHAR(%d),\n", 
							rels[i].attrs[j].name, 
							rels[i].attrs[j].len);
			strcat(buffer, tempbuf);
		}
		strcat(buffer, "\tnl                             CHAR(1)\n");
		strcat(buffer,");\n");

		/* now the insert */
		sprintf(tempbuf, "INSERT INTO %s\nSELECT\n",rels[i].name);
		strcat(buffer, tempbuf);
		for (j = 0; j < rels[i].noattrs; j++) {
			if (rels[i].attrs[j].islenvis == 0) {
				sprintf(tempbuf, "\tcast(%s AS %s)", 
								rels[i].attrs[j].name,
								rels[i].attrs[j].type );
			} else {
				/* 
				 * trim data from the external table if target
				 * relation attribute is VARCHAR or CHAR.
				 * (rtrim is a udf supplied in the Firebird distribution -
				 * altho it needs to be amended to cope with > 80 chars)
				 * otherwise cast as appropriate
				 */
				if ( strcmp(rels[i].attrs[j].type, "VARCHAR") == 0 ||
					 strcmp(rels[i].attrs[j].type, "CHAR") == 0 ){
					sprintf(tempbuf, "\trtrim(%s)", 
									rels[i].attrs[j].name);
				} else {
					sprintf(tempbuf, "\tcast(%s AS %s(%d))", 
									rels[i].attrs[j].name,
									rels[i].attrs[j].type,
									rels[i].attrs[j].len );
				}
			}
			strcat(buffer, tempbuf);
			if (j == rels[i].noattrs - 1 ) {
				strcat(buffer, "\n");
			} else {
				strcat(buffer, ",\n");
			}
		
		}
		sprintf(tempbuf, "FROM %s_ext;\nCOMMIT;\n\n",rels[i].name);
		strcat(buffer, tempbuf);

		/*
		 * write it out
		 */
		if( write(fd, buffer, strlen(buffer)) == -1 ) {
			perror("failed to write");
			return 2;
		}

	}


	close(fd);
	freerels(rels);

	free(tempbuf);
	free(buffer);
	free(lodfile);

	return 0;
}


/*
 * show run options
 */
void	usage() {

	printf("loadgen -t dbtype -d directory -l directory\n");
	printf("explanation of options follows:\n");
	printf("        -d directory : write load files here\n");
	printf("        -l directory : data files are in this directory\n");
	printf("        -t dbtype must be specified, and is one of :\n");
	printf("           pg  - postgresql\n");
	printf("           my  - mysql\n");
	printf("           db2 - db2\n");
	printf("           ora - oracle\n");
	printf("           sap - sapdb\n");
	printf("           fb -  firebird\n");
	return;
}


/*
 * main entry point
 */
int		main(int argc, char** argv) {


	int		option;		

	char	*load_path = ".";
	char	*data_path = ".";
	int		db_type = -1;

	/* 
	 * get and check options 
	 */

	while ((option = getopt(argc, argv, "d:t:l:")) != -1)  {
		switch(option) {
			case 'd':
				data_path = malloc(strlen(optarg));
				strcpy(data_path, optarg);
				break;
			case 'l':
				load_path = malloc(strlen(optarg));
				strcpy(load_path, optarg);
				break;
			case 't':
				if (strcmp(optarg, "pg") == 0){
					db_type = PG;
				} else if (strcmp(optarg, "my") == 0){
					db_type = MY;
				} else if (strcmp(optarg, "db2") == 0){
					db_type = DB2;
				} else if (strcmp(optarg, "ora") == 0){
					db_type = ORA;
				} else if (strcmp(optarg, "sap") == 0){
					db_type = SAP;
				} else if (strcmp(optarg, "fb") == 0) {
					db_type = FB;
				} else {
					usage();
					return 1;
				}
				break;
			default:
				usage();
				return 1;
		}
	}
	argc -= optind;
	argv += optind;

	/*
	 *  catch mandatory option -t
	 */
	if ( db_type == -1 ) {
		usage();
		return 1;
	}


	/*
	 *  get started
	 */
	printf("loadgen : writing load file\n");
	printf("        : for database type %s on path %s\n", 
					db_name(db_type), 
					data_path);
 

	/*
	 *  write load file for whatever database type was selected
	 */
	switch (db_type) {
	case PG :
		if (write_lod_pg(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	case DB2 :
		if (write_lod_db2(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	case MY :
		if (write_lod_my(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	case ORA :
		if (write_lod_ora(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	case SAP :
		if (write_lod_sap(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	case FB :
		if (write_lod_fb(load_path, data_path) != 0) {
			fprintf(stderr, "bailing out during load file write\n");
			return 1;
		}
		break;
	}

	return 0;
}
