/*
 * querygen.c	: generate queries for benchw suite
 *
 * by Mark Kirkwood (markir@paradise.net.nz)
 *
 */


/*
 * required includes
 */
#include "inc.h"


/*
 * relation structures
 */
#include "rel.h"


/*
 * query specific includes
 */
#include "qry.h"


/*
 *
 */
int		write_qry(char *path) {

	Qry		*qrys;
	char	*qryfile;
	char	*buffer;
	char	*tempbuf;
	int		fd;
	int		i, j;
	int		grps, nogrps;


	/*
	 *  allocate enough space for any of the filenames
	 */
	qryfile = malloc(strlen(path) + IDSIZ + 10);


	/*
	 * allocate space for the buffers (again probably overestimating)
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);
	
	
	/*
	 * get the relation structure
	 * and set about doing 1 more query than we have relations
	 */
	qrys = getqrys();
	for ( i = 0 ; i < NOQRYS; i++ ) {
		/*
		 * work out a name and open a file
		 */
		sprintf(qryfile, "%s%s%s%d%s", path, "/", QRYFILE, i, ".sql");
		if ((fd = open(qryfile, OPENOPTS, OPENFLGS )) == -1 ) {
			perror("could not create");
			return 1;
		}
		printf("          :    %s\n", qryfile);


		/*
		 * start writing the buffer
		 *
		 */
		sprintf(buffer,"--\n--  benchw query : generated by querygen\n--               : query type %d\n--\n",i);
		strcat(buffer, "SELECT \n");

		/*
		 * write the projections
		 */
		nogrps = 0;
		for (j = 0; j < qrys[i].noprojs; j++) {
			if ( strcmp(qrys[i].projs[j].funct, "") == 0) {
				sprintf(tempbuf, "\t%s.%s", qrys[i].projs[j].alias, qrys[i].projs[j].attr);
			} else {
				sprintf(tempbuf, "\t%s(%s.%s )", qrys[i].projs[j].funct,
				qrys[i].projs[j].alias, qrys[i].projs[j].attr);
			}
			if (j != qrys[i].noprojs - 1) {
				strcat(tempbuf, ",\n");
			} else {
				strcat(tempbuf, "\n");
			}
			if (qrys[i].projs[j].isagg == 0) {
				nogrps++;
			}
			strcat(buffer, tempbuf);
		}


		/*
		 * write the relations
		 */
		strcat(buffer, "FROM\n");
		for (j = 0; j < qrys[i].norels; j ++) {
			sprintf(tempbuf, "\t%s AS %s", qrys[i].rels[j].name, qrys[i].rels[j].alias);
			if (j != qrys[i].norels - 1) {
				strcat(tempbuf, ",\n");
			} else {
				strcat(tempbuf,"\n");
			}
			strcat(buffer, tempbuf);
		}
	
		

		/*
		 * write the joins
		 */
		for (j = 0; j < qrys[i].norels - 1; j++) {
			if ( j == 0) {
				sprintf(tempbuf, "WHERE \t%s.%s = %s.%s\n",
						qrys[i].joins[j].fromalias, qrys[i].joins[j].fromattr,
						qrys[i].joins[j].toalias, qrys[i].joins[j].toattr);
			} else {
				sprintf(tempbuf, "AND \t%s.%s = %s.%s\n",
						qrys[i].joins[j].fromalias, qrys[i].joins[j].fromattr,
						qrys[i].joins[j].toalias, qrys[i].joins[j].toattr);
			}
			strcat(buffer, tempbuf);
		}


		/*
		 * write the remaining restrictions
		 */
		for (j = 0; j < qrys[i].norests ; j++) {
			sprintf(tempbuf, "AND \t%s.%s %s %s\n",
					qrys[i].rests[j].alias, qrys[i].rests[j].attr,
					qrys[i].rests[j].op, qrys[i].rests[j].val);
			strcat(buffer, tempbuf);
		}


		/*
		 *  group by the non-aggregate projections
		 */
		grps = 0;
		if ( nogrps > 0 ) {
			strcat(buffer, "GROUP BY\n");
		}
		
		for (j = 0; j < qrys[i].noprojs; j++) {
			if ( qrys[i].projs[j].isagg == 0) {
				sprintf(tempbuf, "\t%s.%s",
						qrys[i].projs[j].alias, qrys[i].projs[j].attr);
				grps++;
				if (grps < nogrps) {
					strcat(tempbuf, ",\n");
				} else {
					strcat(tempbuf, "\n");
				}
				strcat(buffer, tempbuf);
			}
			
		}


		/*
		 * finally append the sql terminator
		 */
		strcat(buffer, ";\n");


		/*
		 * write it out and then close
		 */
		if( write(fd, buffer, strlen(buffer)) == -1 ) {
			perror("failed to write");
			return 2;
		}

		close(fd);

	}

	freeqrys(qrys);
	
	free(tempbuf);
	free(buffer);

	return 0;
	
}


/*
 * show run options
 */
void	usage() {

	printf("querygen -d directory\n");
	return;
}


/*
 * main entry point
 */
int		main(int argc, char** argv) {


	int		option;

	char	*data_path = ".";

	/*
	 * get and check options
	 */

	while ((option = getopt(argc, argv, "d:")) != -1)  {
		switch(option) {
			case 'd':
				data_path = malloc(strlen(optarg));
				strcpy(data_path, optarg);
				break;
			default:
				usage(argv[0]);
				return 1;
		}
	}
	argc -= optind;
	argv += optind;

	/*
	 * get started
	 */
	printf("querygen  : writing query files\n");
	printf("          : on path %s\n", data_path);

	if (write_qry(data_path) != 0) {
		fprintf(stderr, "bailing out during query file write\n");
		return 1;
	}

	return 0;
}
