/*
 * schemagen.c	: generate schema for benchw suite
 * 
 * by Mark Kirkwood (markir@paradise.net.nz)
 *
 * Bugs 		: should really check return code from all malloc's
 */


/*
 * required includes
 */
#include "inc.h"


/*
 * relation structures
 */
#include "rel.h"

/*
 * schema specific defines etc
 */
#include "sch.h"


/*
 * write the schema creation file
 */
int	write_sch(char *path) {


	char	*schfile;
	int		fd;
	char	*buffer;
	char	*tempbuf;
	int 	i,j;
	Rel		*rels;


	/*
	 * work out the file name
	 */
	schfile = malloc(strlen(path) + strlen(SCHFILE) + 2);
	sprintf(schfile, "%s%s%s", path, "/", SCHFILE);


	/*
	 * open the file
	 */
	if ((fd = open(schfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("          :    %s\n", schfile);


	/*
	 * give the buffer some space (very wet finger here... )
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);


	/*
	 * start writing the buffer out
	 *
	 */
	sprintf(buffer, "--\n--  benchw schema : generated by schemagen\n--\n");


	/*
	 * get the relation structures
	 * then write out the create table statements
	 * complete with attributes
	 */
	rels = getrels();
	for (i = 0; i < NORELS; i++) {
		sprintf(tempbuf, "\nCREATE TABLE %s (\n", rels[i].name);
		strcat(buffer, tempbuf);

		for (j = 0; j < rels[i].noattrs; j++) {
			if ( rels[i].attrs[j].islenvis == 0 ) {
				sprintf(tempbuf, "\t%-30s %s NOT NULL",
						rels[i].attrs[j].name, rels[i].attrs[j].type);
			} else {
				sprintf(tempbuf, "\t%-30s %s(%d) NOT NULL",
						rels[i].attrs[j].name, rels[i].attrs[j].type,
						rels[i].attrs[j].len);
			}
			if ( j != rels[i].noattrs -1 ) {		// comma if not at end of attrs !
				strcat(tempbuf, ",");
			}
			strcat(tempbuf, "\n");
			strcat(buffer, tempbuf);
		}
		strcat(buffer, ");\n");
	}


	/*
	 * write out the buffer
	 */
	if( write(fd, buffer, strlen(buffer)) == -1 ) {
		perror("failed to write");
		return 2 ;
	}

	
	freerels(rels);
	
	free(buffer);
	free(tempbuf);
	close(fd);
	free(schfile);

	return 0;
}


/*
 * write default indexes creation file
 */
int	write_ind(char *path) {


	char	*indfile;
	int		fd;
	char	*buffer;
	char	*tempbuf;
	int 	i,j;
	Rel		*rels;


	/*
	 * work out the file name
	 */
	indfile = malloc(strlen(path) + strlen(INDFILE) + 2);
	sprintf(indfile, "%s%s%s", path, "/", INDFILE);


	/*
	 * open the file
	 */
	if ((fd = open(indfile, OPENOPTS, OPENFLGS )) == -1 ) {
		perror("could not create");
		return 1;
	}
	printf("          :    %s\n", indfile);



	/*
	 * give the buffer some space (very wet finger here... )
	 */
	buffer = malloc(4096);
	tempbuf = malloc(512);


	/*
	 * start writing the buffer out
	 *
	 */
	sprintf(buffer, "--\n--  benchw (suggested) indexes : generated by schemagen\n--\n");


	
	/*
	 * get the relation structures,
	 * examine the attributes for each one, to see if they
	 * are suitable for indexing
	 */
	rels = getrels();
	for (i = 0; i < NORELS; i++) {
		for (j = 0; j < rels[i].noattrs; j++) {
			if (rels[i].attrs[j].iskey == 1) {
				
				if (strcmp(rels[i].type,"fact") == 0) {
					strcat(buffer, "CREATE INDEX ");
				} else {
					strcat(buffer, "CREATE UNIQUE INDEX ");
				}
				
				sprintf(tempbuf, "%s_%s ON %s(%s);\n\n",
						rels[i].name, rels[i].attrs[j].name,
						rels[i].name, rels[i].attrs[j].name);
				strcat(buffer, tempbuf);
			}
		}
	}

	
	/*
	 * write out the buffer
	 */
	if( write(fd, buffer, strlen(buffer)) == -1 ) {
		perror("failed to write");
		return 2 ;
	}

	
	freerels(rels);
	free(buffer);
	free(tempbuf);
	close(fd);
	free(indfile);

	return 0;
}


/*
 * show run options
 */
void	usage() {

	printf("schemagen -d directory\n");
	return;
}


/*
 * main entry point
 */
int		main(int argc, char** argv) {


	int		option;		

	char	*data_path = ".";

	/* 
	 * get and check options 
	 */

	while ((option = getopt(argc, argv, "d:")) != -1)  {
		switch(option) {
			case 'd':
				data_path = malloc(strlen(optarg));
				strcpy(data_path, optarg);
				break;
			default:
				usage(argv[0]);
				return 1;
		}
	}
	argc -= optind;
	argv += optind;

	/*
	 * get started, write a schema creation file
	 * write a suggested index file seperately
	 */
	printf("schemagen : writing schema and default index creation files\n");
	printf("          : on path %s\n", data_path);

	if (write_sch(data_path) != 0) {
		fprintf(stderr, "bailing out during write schema\n");
		return 1;
	}


	if (write_ind(data_path) != 0) {
		fprintf(stderr, "bailing out during write indexes\n");
		return 1;
	}


	return 0;
}
